#define bqt_unicode_cc

#include <cstdio>
#include "unicode/unicode.hh"

static const struct { const char *ent; wchar_t ch; } EntityTable[] =
{
#include "entities.h"
};
static const wchar_t FindEntity(const string &s)
{
    unsigned size = sizeof(EntityTable) / sizeof(EntityTable[0]);
    unsigned eka=0, vika=size-1;
    for(;;)
    {
        unsigned n = (eka+vika)/2;
        if(s == EntityTable[n].ent)return EntityTable[n].ch;
        if(eka >= vika)break;
        if(s > EntityTable[n].ent) { eka=n+1; continue; }
        if(s < EntityTable[n].ent) { vika=n; continue; }
    }
    return '?';
}

wstring charset::TryParseEntities(const string &s, unsigned &a) return ss
{
    while(a < s.size() && s[a]=='&')
    {
        ++a;
        unsigned c = s.find(';', a);
        if(c != s.npos)  
        {
            string code = s.substr(a, c-a);
            a = c+1;
            wchar_t ch = '?';
            if(code.size() > 0)
            {
                if(code[0] == '#')
                {
                    if(code.size() > 1 && (code[1] == 'x' || code[1] == 'X'))
                        ch = strtol(code.c_str()+2, NULL, 16);
                    else
                        ch = strtol(code.c_str()+1, NULL, 10);
                }
                else
                {
                    ch = FindEntity(code);
                }
            }
            ss += ch;
        }
    }
}

class easycharset : public charset
{
    const charmap *in;
    const charmap *out;
public:
    easycharset(const charmap *i, const charmap *o) : in(i), out(o) { }
	virtual const wstring c2uni(const string &s, bool html=false)
    {
    	wstring ss;
        unsigned a, b=s.size();
        for(a=0; a<b; ++a)
        {
        	if(html)
        	{
    	    	ss += TryParseEntities(s, a);
        		if(a==b)break;
        	}
        	charmap::const_iterator i = in->find(s[a]);
        	if(i == in->end())
        		ss += 0xFFFD; /* undefined char */
        	else
    	        ss += i->second;
        }
        return ss;
    }
	virtual const string uni2c(const wstring &s)
    {
    	string ss;
        unsigned a, b=s.size();
        for(a=0; a<b; ++a)
        {
        	charmap::const_iterator i = out->find(s[a]);
        	if(i == out->end())
        		ss += '?';
        	else
    	        ss += (char)i->second;
        }
        return ss;
    }
};

class jis208 : public charset
{
	enum jistate { ascii, st201,
	               st78a, st78b,
	               st83a, st83b,
	               gotesc, gotsulku, gotdolla };
	jistate instate, outstate;
	unsigned char cache;
public:
    jis208() : instate(ascii), outstate(ascii)
    {
    }
	virtual const wstring c2uni(const string &s, bool html=false)
	{
		wstring ss;
		unsigned a, b=s.size();
		for(a=0; a<b; ++a)
		{
			switch(instate)
			{
		    	case gotesc:
		    		if(s[a]=='(') { instate = gotsulku; break; }
		    		if(s[a]=='$') { instate = gotdolla; break; }
		    		instate = ascii; 
		    		/* fallthru */
				case ascii:
		        	if(html)
		        	{
		    	    	ss += TryParseEntities(s, a);
		        		if(a==b)break;
		        	}
					if(s[a]=='\33') { instate = gotesc; break; }
					ss += s[a]; /* unconverted - assume iso-8859-1 */
		    	    break;
		    	case gotsulku:
		    		instate = ascii;
		    		if(s[a]=='J') instate = st201;
		    		break;
		    	case gotdolla:
		    		instate = ascii;
		    		if(s[a]=='@') instate = st78a;
		    		else if(s[a]=='B') instate = st83a;
		    		break;
		    	case st201:
		    	{
		    		if(s[a] == '\33') { instate = gotesc; break; }
		        	charmap::const_iterator i = jis201_to_uni.find(s[a]);
		        	if(i == jis201_to_uni.end())
		        		ss += 0xFFFD; /* undefined char */
        			else
		    	        ss += i->second;
		    		break;
		    	}
		    	case st78a:
		    	case st78b:
		    		if(s[a] == '\33') { instate = gotesc; break; }
		    		/* unknown */
		    		break;
		    	case st83a:
		    		if(s[a] == '\33') { instate = gotesc; break; }
		    		cache = s[a];
		    		instate = st83b;
		    		break;
		    	case st83b:
		    	{
		    		if(s[a] == '\33') { instate = gotesc; break; }
		    		wchar_t t = cache*256 + s[a];
		    		// fprintf(stderr, "Finding %04X (%c%c)\n", t, cache, s[a]);
		        	charmap::const_iterator i = jis208_to_uni.find(t);
		        	if(i == jis208_to_uni.end())
		        		ss += 0xFFFD; /* undefined char */
        			else
		    	        ss += i->second;
		    	    instate = st83a;
		    		break;
		    	}
			}
		}
		return ss;
	}
	virtual const string uni2c(const wstring &s)
	{
		string ss;
		unsigned a, b=s.size();
		for(a=0; a<b; ++a)
		{
        	charmap::const_iterator i = uni_to_jis208.find(s[a]);
        	if(i == uni_to_jis208.end())
        	{
	        	i = uni_to_jis201.find(s[a]);
	        	if(i == uni_to_jis201.end())
	        	{
	        		if(outstate != ascii) { ss += "\33(B"; outstate = ascii; }
    	    		if(s[a] < 0x100)
	    	    		ss += (char)s[a];
	        		else
	        			ss += '?';
	        	}
	        	else
	        	{
	        		if(outstate != st201) { ss += "\33(J"; outstate = st201; }
	        		ss += (char)i->second;
	        	}
       		}
        	else
        	{
        		if(outstate != st83a) { ss += "\33$B"; outstate = st83a; }
    	        ss += (char)(i->second >> 8);
    	        ss += (char)(i->second & 127);
    	    }
		}
		/* Always end with ascii mode */
		if(outstate != ascii) { ss += "\33(B"; outstate = ascii; }
		return ss;
	}
};

class shift_jis : public charset
{
	unsigned char cache;
public:
	shift_jis() : cache(0)
	{
	}
	virtual const wstring c2uni(const string &s, bool html=false)
    {
    	wstring ss;
        unsigned a, b=s.size();
        for(a=0; a<b; ++a)
        {
        	wchar_t t;
        	if(cache)
        	{
        		t = cache*256 + s[a];
        	Redo:
	        	charmap::const_iterator i = sjis_to_uni.find(t);
	        	if(i == sjis_to_uni.end())
	        		ss += 0xFFFD; /* undefined char */
       			else
	    	        ss += i->second;
	    	    cache = 0;
        	}
        	else
        	{
	        	if(html)
	        	{
	    	    	ss += TryParseEntities(s, a);
	        		if(a==b)break;
	        	}
        		unsigned char cache = s[a];
        		if(cache < 0x80 || (cache>=0xA0 && cache<0xE0))
        		{
        			t = cache;
        			cache = 0;
        			goto Redo;
        		}
        	}
        }
        return ss;
    }
	virtual const string uni2c(const wstring &s)
	{
    	string ss;
        unsigned a, b=s.size();
        for(a=0; a<b; ++a)
        {
        	charmap::const_iterator i = uni_to_sjis.find(s[a]);
        	if(i == uni_to_sjis.end())
        	{
        		if((unsigned char)s[a] < 32)
        			ss += (char)s[a];
        		else
        			ss += '?';
        	}
        	else if(i->second >= 0x100)
        	{
        		ss += (char)(i->second >> 8);
        		ss += (char)(i->second & 255);
        	}
       		else
   	        	ss += (char)i->second;
        }
        return ss;
	}
};

class utf8 : public charset
{
	unsigned bytesleft, cache;
public:
	utf8() : bytesleft(0), cache(0) { }
	virtual const wstring c2uni(const string &s, bool html=false)
	{
    	wstring ss;
        unsigned a, b=s.size();
        for(a=0; a<b; ++a)
        {
        	unsigned char c = s[a];
        	if(bytesleft > 0) { --bytesleft; cache = cache * 64 + (c&63); }
        	else
        	{
	        Redo:
        		if((c & 0xE0) == 0xC0) { bytesleft=1; cache = c&0x1F; }
        		else if((c & 0xF0) == 0xE0) { bytesleft=2; cache = c&0x0F; }
        		else if((c & 0xF8) == 0xF0) { bytesleft=3; cache = c&0x07; }
        		else
        		{
		        	if(html)
		        	{
	    		    	ss += TryParseEntities(s, a);
	        			if(a==b)break;
	        			c = s[a];
	        			if(c&0x80)goto Redo;
	        		}
        			cache = c&0x7F;
        		}
        	}
        	if(!bytesleft)
        	{
        		ss += (wchar_t)cache;
        	}
        }
        return ss;
	}
	virtual const string uni2c(const wstring &s)
	{
    	string ss = "\33%G"; /* Starts UTF-8 in Linux */
        unsigned a, b=s.size();
        for(a=0; a<b; ++a)
        {
        	wchar_t c = s[a];
        	if(c < 0x80)
        		ss += (unsigned char)c;
        	else if(c < 0x800)
        	{
        		ss += (unsigned char)(0xC0 + (c>>6));
        		ss += (unsigned char)(0x80 + (c&63));
        	}
        	else if(c < 0x10000)
        	{
        		ss += (unsigned char)(0xE0 + (c>>12));
        		ss += (unsigned char)(0x80 + ((c>>6)&63));
        		ss += (unsigned char)(0x80 + (c&63));
        	}
        	else
        	{
        		ss += (unsigned char)(0xE0 + (c>>18));
        		ss += (unsigned char)(0x80 + ((c>>12)&63));
        		ss += (unsigned char)(0x80 + ((c>>6)&63));
        		ss += (unsigned char)(0x80 + (c&63));
        	}
        }
        ss += "\33%@";     /* Ends UTF-8 in Linux */
        return ss;
	}
};

charset *newconverter(const string &csetname)
{
#include "unicode/converters.hh"	
    if(csetname=="iso-2022-jp")return new jis208();
    if(csetname=="shift_jis")return new shift_jis();
    if(csetname=="utf8")return new utf8();
	return new easycharset(&iso01_to_uni, &uni_to_iso01);
}
